#!/usr/bin/env python3
# encoding: utf-8
#
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2024 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2024 Henrik tom Wörden <h.tomwoerden@indiscale.com>
# Copyright (C) 2024 Daniel Hornung <d.hornung@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import json
from types import SimpleNamespace
from typing import List, Union, TextIO

from openpyxl import load_workbook

from .table_generator import ColumnType, RowType


def _fill_leaves(json_doc: dict, workbook):
    for key, value in json_doc:
        if not isinstance(value, list):
            value = [value]
        for el in value:
            if isinstance(el, dict):
                _fill_leaves(el, workbook)
            workbook.cell(1, 2, el)


def _get_row_type_column_index(worksheet):
    for col in worksheet.columns:
        for cell in col:
            if cell.value == RowType.COL_TYPE.name:
                return cell.column
    raise ValueError("The column which defines row types (COL_TYPE, PATH, ...) is missing")


def _get_path_rows(worksheet):
    rows = []
    rt_col = _get_row_type_column_index(worksheet)
    for cell in list(worksheet.columns)[rt_col-1]:
        print(cell.value)
        if cell.value == RowType.PATH.name:
            rows.append(cell.row)
    return rows


def _next_row_index(sheet) -> int:
    """Return the index for the next data row.

This is defined as the first row without any content.
"""
    return sheet.max_row


class TemplateFiller:
    def __init__(self, workbook):
        self._workbook = workbook
        self._create_index()

    def fill_data(self, data: dict):
        """Fill the data into the workbook."""
        self._handle_data(data=data, current_path=[])

    def _create_index(self, ):
        """Create a sheet index for the workbook.

        Index the sheets by their relevant path array.  Also create a simple column index by column
        type and path.

        """
        self._sheet_index = {}
        for sheetname in self._workbook.sheetnames:
            sheet = self._workbook[sheetname]
            type_column = [x.value for x in list(sheet.columns)[
                _get_row_type_column_index(sheet) - 1]]
            # 0-indexed, as everything outside of sheet.cell(...):
            coltype_idx = type_column.index(RowType.COL_TYPE.name)
            path_indices = [i for i, typ in enumerate(type_column) if typ == RowType.PATH.name]

            # Get the paths, use without the leaf component for sheet indexing, with type prefix and
            # leaf for column indexing.
            paths = []
            col_index = {}
            for col_idx, col in enumerate(sheet.columns):
                if col[coltype_idx].value == RowType.COL_TYPE.name:
                    continue
                path = []
                for path_idx in path_indices:
                    if col[path_idx].value is not None:
                        path.append(col[path_idx].value)
                col_key = ".".join([col[coltype_idx].value] + path)
                col_index[col_key] = SimpleNamespace(column=col, col_index=col_idx)
                if col[coltype_idx].value not in [ColumnType.SCALAR.name, ColumnType.LIST.name]:
                    continue
                paths.append(path[:-1])

            # Find common components:
            common_path = []
            for idx, component in enumerate(paths[0]):
                for path in paths:
                    if not path[idx] == component:
                        break
                else:
                    common_path.append(component)
            assert len(common_path) >= 1

            self._sheet_index[".".join(common_path)] = SimpleNamespace(
                common_path=common_path, sheetname=sheetname, sheet=sheet, col_index=col_index)

    def _handle_data(self, data: dict, current_path: List[str] = None):
        """Handle the data and write it into ``workbook``.

Parameters
----------
data: dict
  The data at the current path position.  Elements may be dicts, lists or simple scalar values.
        """
        if current_path is None:
            current_path = []
        for name, content in data.items():
            path = current_path + [name]
            if isinstance(content, list):
                if not content:
                    continue
                # Must be all of the same type.
                assert len(set(type(entry) for entry in content)) == 1
                if isinstance(content[0], dict):
                    # An array of objects: must go into exploded sheet
                    for entry in content:
                        self._handle_data(data=entry, current_path=path)
                    continue
            self._handle_simple_data(data=content, current_path=path)

    def _handle_simple_data(self, data, current_path: List[str]):
        """Enter this single data item into the workbook.

Parameters
----------
data: dict
  The data at the current path position.  Must be single items (dict or simple scalar) or lists of
  simple values.
        """
        sheet = self._sheet_index[".".join(current_path)].sheet
        next_row = _next_row_index(sheet)
        for name, content in data.items():
            if isinstance(content, list):
                # TODO handle later
                # scalar elements: semicolon separated
                # nested dicts: recurse
                continue
            if isinstance(content, dict):
                pass
                # from IPython import embed
                # embed()


def fill_template(data: Union[dict, str, TextIO], template: str, result: str) -> None:
    """Insert json data into an xlsx file, according to a template.

This function fills the json data into the template stored at ``template`` and stores the result as
``result``.

Parameters
----------
data: Union[dict, str, TextIO]
  The data, given as Python dict, path to a file or a file-like object.
template: str
  Path to the XLSX template.
result: str
  Path for the result XLSX.
"""
    if isinstance(data, dict):
        pass
    elif isinstance(data, str):
        with open(data, encoding="utf-8") as infile:
            data = json.load(infile)
    elif hasattr(data, "read"):
        data = json.load(data)
    else:
        raise ValueError(f"I don't know how to handle the datatype of `data`: {type(data)}")
    result_wb = load_workbook(template)
    template_filler = TemplateFiller(result_wb)
    # For each top level key in the json we iterate the values (if it is an array). Those are the
    # root elements that belong to a particular sheet.
    #       After treating a root element, the row index for the corresponding sheet needs to be
    #       increased
    #       When we finished treating an object that goes into a lower ranked sheet (see below), we
    #       increase the row index of that sheet.
    #
    # We can generate a hierarchy of sheets in the beginning (using the paths). The lower sheets
    # are for objects referenced by objects in higher ranked sheets.
    # We can detect the sheet corresponding to a root element by looking at the first path element:
    # The first path element must be the root element every where.
    # Suggestion:
    # row indices: Dict[str, int] string is the sheet name
    # sheet_hirarchy: List[Tuple[str]]  elements are sheet names
    template_filler.fill_data(data=data)

    # Question:
    # We can create an internal representation where we assign as sheet_names the same names that
    # are used in table generator. Or should we create another special row that contains this
    # somehow?

    result_wb.save(result)
